//
// $Id: RADIUShelpers.cpp,v 1.1 2004/01/27 16:20:31 holger Exp holger $
//
// RADIUS Plugin for pGina
// Copyright (c) 2004 Holger Weiss
//
// ---------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc., 59
// Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// ---------------------------------------------------------------------------
//
// The UNICODE conversion functions are slightly modified versions of similar
// functions taken from the PAM Plugin for pGina, copyright (c) 2003 Nathan
// Yocom.  The logging function, getWindowsVersion and the Registry functions
// are based on functions taken from pGina, copyright (c) 2002 Nathan Yocom.
// Some of the IP related code is taken from pam_radius_auth, copyright (c)
// 2003 Alan DeKok.  The pseudorandom integer algorithm is copyright (c) 1997
// Makoto Matsumoto and Takuji Nishimura (their code is LGPL'd, not GPL'd).
//

#include "stdafx.h"
#include "RADIUShelpers.h"

#include <windows.h>
#include <winsock2.h>
#include <lm.h>
#include <mmsystem.h>
#include <stdarg.h>
#include <stdio.h>
#include <tchar.h>

// ---------------------------------------------------------------------------
// Pseudorandom 32bit integers

// genrand() generates one pseudorandom unsigned integer (32bit) which is
// uniformly distributed among 0 to 2^32-1 for each call.  sgenrand(seed) sets
// initial values to the working area of 624 words.  Before genrand() is
// called, sgenrand(seed) must be called once (seed is any 32-bit integer
// except for 0).

// Initializing the array with a NONZERO seed

void sgenrand(unsigned long seed)
{
	// setting initial seeds to mt[N] using the generator Line 25 of Table
	// 1 in [KNUTH 1981, The Art of Computer Programming Vol. 2 (2nd Ed.),
	// pp102]

	mt[0]= seed & 0xffffffff;
	for (mti=1; mti<N; mti++)
		mt[mti] = (69069 * mt[mti-1]) & 0xffffffff;
}

unsigned long genrand(void)
{
	unsigned long y;
	static unsigned long mag01[2]={0x0, MATRIX_A};

	if (mti >= N) {	// generate N words at one time
		int kk;

		if (mti == N+1)		// if sgenrand() has not been called,
			sgenrand(4357);	// a default initial seed is used

		for (kk=0;kk<N-M;kk++) {
			y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
			mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1];
		}
		for (;kk<N-1;kk++) {
			y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
			mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1];
		}
		y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
		mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1];

		mti = 0;
	}

	y = mt[mti++];
	y ^= TEMPERING_SHIFT_U(y);
	y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
	y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
	y ^= TEMPERING_SHIFT_L(y);

	return y;
}

// ---------------------------------------------------------------------------
// Emulate UNIX' gettimeofday

int gettimeofday(struct timeval *time_Info, struct timezone *timezone_Info)
{
	// get the time, if they want it
	if (time_Info != NULL) {
		time_Info->tv_sec = (long)time(NULL);
		time_Info->tv_usec = timeGetTime()*1000;
	}

	// get the timezone, if they want it
	if (timezone_Info != NULL) {
		_tzset();
		timezone_Info->tz_minuteswest = _timezone;
		timezone_Info->tz_dsttime = _daylight;
	}

	return 0;
}

// ---------------------------------------------------------------------------
// UNICODE conversion

char *convert_to_char(const TCHAR *from)
{
	char *to = (char *) GlobalAlloc(NULL, BUFFER_SIZE);

	WideCharToMultiByte(CP_ACP,
	                    NULL,
	                    (LPCWSTR)from,
	                    -1,
	                    to,
	                    BUFFER_SIZE,
	                    NULL,
	                    NULL);

	return to;
}

TCHAR *convert_to_wchar(const char *from)
{
	TCHAR *to;

	to = (TCHAR *) GlobalAlloc(NULL, BUFFER_SIZE*sizeof(TCHAR)+1);

	MultiByteToWideChar(CP_ACP,
	                    NULL,
	                    (LPCSTR)from,
	                    (int)strlen(from)+1,
	                    (LPWSTR)to,
	                    BUFFER_SIZE*sizeof(TCHAR)+1);

	return to;
}

// ---------------------------------------------------------------------------
// Windows Version / Logging

int getWindowsVersion(void)
{
	OSVERSIONINFOEX osvi;
	BOOL bOsVersionInfoEx;

	// Try calling GetVersionEx using the OSVERSIONINFOEX structure.
	// If that fails, try using the OSVERSIONINFO structure.

	ZeroMemory(&osvi, sizeof(OSVERSIONINFOEX));
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

	if (!(bOsVersionInfoEx = GetVersionEx((OSVERSIONINFO *) &osvi))) {
		osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		if (!GetVersionEx((OSVERSIONINFO *)&osvi))
			return WINDOWS_NT_4;	// default to NT 4
	}

	switch (osvi.dwPlatformId) {

	// test for the Windows NT product family
	case VER_PLATFORM_WIN32_NT:

		// test for the specific product family
		if (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2)
			return WINDOWS_2003;

		if (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 1)
			return WINDOWS_XP;

		if (osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0)
			return WINDOWS_2000;

		if (osvi.dwMajorVersion <= 4)
			return WINDOWS_NT_4;

		break;

	// test for the Windows 95 product family.
	case VER_PLATFORM_WIN32_WINDOWS:
		return WINDOWS_SUCKY;
		break;
	}

	return WINDOWS_NT_4;	// default to NT 4
}

void radiusLog(WORD log_type, char *string, ...)
{
	LPCSTR *strings[1];
	char buffer[30000];	// max log string size
	va_list argList;
	FILE *debugFile;
	HANDLE h;
	static int WindowsVersion = getWindowsVersion();

	if (WindowsVersion != WINDOWS_NT_4) {

		va_start(argList, string);	// setup the string to print
		vsprintf((char *)buffer, (const char *)string, argList);
		strings[0] = (LPCSTR *)&buffer;

		// this saves to "Application" if "RADIUSplugin" EventSource doesn't exist
		h = RegisterEventSourceA(NULL, "RADIUSplugin");
		ReportEventA(h, log_type, 0, 0, NULL, 1, 0, (LPCSTR *)strings, NULL);
		DeregisterEventSource(h);

	} else {

		// on NT, simply save to C:\RADIUSplugin_log.txt
		if (debugFile = fopen("c:\\RADIUSplugin_log.txt", "a")) {

			va_start(argList, string);	// setup the string to print
			vsprintf((char *)buffer, (const char *)string, argList);
			fprintf(debugFile, "%s\n", buffer);
			fflush(debugFile);
			fclose(debugFile);
		}
	}
}

// ---------------------------------------------------------------------------
// Registry

// We use the non UNICODE functions no matter whether or not UNICODE is
// defined (*urgs*)

char *regReadString(LPCSTR location)
{
	LONG lResult;
	HKEY hKey;
	DWORD dwBytesRead = 2048;
	char result[2048];
	char *ret;

	lResult = RegOpenKeyExA(HKEY_LOCAL_MACHINE,
	                        "SOFTWARE\\pGina\\RADIUSplugin",
	                        0,
	                        KEY_QUERY_VALUE,
	                        &hKey);

	if (lResult == ERROR_SUCCESS) {
		lResult = RegQueryValueExA(hKey,
		                           location,
		                           NULL,
		                           NULL,
		                           (LPBYTE)&result,
		                           &dwBytesRead);

		if (lResult == ERROR_SUCCESS) {
			if (dwBytesRead != 0) {
				ret = strdup(result);
				RegCloseKey(hKey);
				return ret;
			} else {
				return NULL;
			}
		}
	}

	RegCloseKey(hKey);
	return NULL;
}

unsigned long regReadUlong(LPCSTR location)
{
	LONG lResult;
	HKEY hKey;
	DWORD value = 0;
	DWORD size = sizeof(DWORD);

	lResult = RegOpenKeyExA(HKEY_LOCAL_MACHINE,
	                        "SOFTWARE\\pGina\\RADIUSplugin",
	                        0,
	                        KEY_QUERY_VALUE,
	                        &hKey);

	if (lResult == ERROR_SUCCESS) {
		lResult = RegQueryValueExA(hKey,
		                           location,
		                           NULL,
		                           NULL,
		                           (LPBYTE)&value,
		                           &size);

		if (lResult == ERROR_SUCCESS) {
			RegCloseKey(hKey);
			return value;
		}
	}

	RegCloseKey(hKey);
	return 0;
}

BOOL regReadBool(LPCSTR location)
{
	LONG lResult;
	HKEY hKey;
	DWORD value = 0;
	DWORD size = sizeof(DWORD);

	lResult = RegOpenKeyExA(HKEY_LOCAL_MACHINE,
	                        "SOFTWARE\\pGina\\RADIUSplugin",
	                        0,
	                        KEY_QUERY_VALUE,
	                        &hKey);

	if (lResult == ERROR_SUCCESS) {
		lResult = RegQueryValueExA(hKey,
		                           location,
		                           NULL,
		                           NULL,
		                           (LPBYTE)&value,
		                           &size);

		if (lResult == ERROR_SUCCESS) {
			RegCloseKey(hKey);
			if (value == 1)
				return true;
			else
				return false;
		}
	}

	RegCloseKey(hKey);
	return false;
}

BOOL regWriteString(LPCSTR value, LPCSTR location)
{
	HKEY hKey;

	if (RegCreateKeyEx(HKEY_LOCAL_MACHINE,
	                   _T("SOFTWARE\\pGina\\RADIUSplugin"),
	                   0,
	                   NULL,
	                   REG_OPTION_NON_VOLATILE,
	                   KEY_ALL_ACCESS,
	                   NULL,
	                   &hKey,
	                   NULL) == ERROR_SUCCESS) {
		if (value != NULL)
			if (RegSetValueExA(hKey,
			                   location,
			                   0,
			                   REG_SZ,
			                   (CONST BYTE *) value,
			                   (int)strlen(value)*sizeof(char)) == ERROR_SUCCESS) {
				RegFlushKey(hKey);
				RegCloseKey(hKey);
				return true;
			}
	}

	RegFlushKey(hKey);
	RegCloseKey(hKey);
	return false;
}

BOOL regWriteUlong(unsigned long value, LPCSTR location)
{
	HKEY hKey;

	if (RegCreateKeyEx(HKEY_LOCAL_MACHINE,
	                   _T("SOFTWARE\\pGina\\RADIUSplugin"),
	                   0,
	                   NULL,
	                   REG_OPTION_NON_VOLATILE,
	                   KEY_ALL_ACCESS,
	                   NULL,
	                   &hKey,
	                   NULL) == ERROR_SUCCESS)
		if (RegSetValueExA(hKey,
		                   location,
		                   0,
		                   REG_DWORD,
		                   (CONST BYTE *)&value,
		                   sizeof(DWORD)) == ERROR_SUCCESS) {
			RegFlushKey(hKey);
			RegCloseKey(hKey);
			return true;
		}

	RegFlushKey(hKey);
	RegCloseKey(hKey);
	return false;
}

BOOL regWriteBool(BOOL value, LPCSTR location)
{
	HKEY hKey;
	DWORD write;

	write = (value ? 1 : 0);

	if (RegCreateKeyEx(HKEY_LOCAL_MACHINE,
	                   _T("SOFTWARE\\pGina\\RADIUSplugin"),
	                   0,
	                   NULL,
	                   REG_OPTION_NON_VOLATILE,
	                   KEY_ALL_ACCESS,
	                   NULL,
	                   &hKey,
	                   NULL) == ERROR_SUCCESS)
		if (RegSetValueExA(hKey,
		                   location,
		                   0,
		                   REG_DWORD,
		                   (LPBYTE)&write,
		                   sizeof(DWORD)) == ERROR_SUCCESS) {
			RegFlushKey(hKey);
			RegCloseKey(hKey);
			return true;
		}

	RegFlushKey(hKey);
	RegCloseKey(hKey);
	return false;
}

BOOL regSecureKey(void)
{
	SECURITY_ATTRIBUTES sa;
	SECURITY_DESCRIPTOR sd;
	BYTE aclBuffer[BUFFER_SIZE];
	PACL pacl=(PACL)&aclBuffer;
	PSID psid;
	SID_IDENTIFIER_AUTHORITY siaNtAuthority = SECURITY_NT_AUTHORITY;
	HKEY hKey;

	if (!InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION)
	        || !InitializeAcl(pacl, sizeof(aclBuffer), ACL_REVISION)
	        || !AllocateAndInitializeSid(&siaNtAuthority,
	                                     2,
	                                     SECURITY_BUILTIN_DOMAIN_RID,
	                                     DOMAIN_ALIAS_RID_ADMINS,
	                                     0, 0, 0, 0, 0, 0,
	                                     &psid) // psid now represents the
	                                            // local administrators group
	        || !AddAccessAllowedAce(pacl, ACL_REVISION, GENERIC_ALL, psid)
	        || !SetSecurityDescriptorDacl(&sd, TRUE, pacl, FALSE))
		return false;

	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.bInheritHandle = FALSE;
	sa.lpSecurityDescriptor = &sd;

	if (RegOpenKeyExA(HKEY_LOCAL_MACHINE,
	                  "SOFTWARE\\pGina\\RADIUSplugin",
	                  0,
	                  WRITE_DAC,
	                  &hKey) != ERROR_SUCCESS)
		return false;

	if (RegSetKeySecurity(hKey,
	                      (SECURITY_INFORMATION)DACL_SECURITY_INFORMATION,
	                      &sd) != ERROR_SUCCESS)
		return false;

	RegFlushKey(hKey);
	RegCloseKey(hKey);
	FreeSid(psid);

	return true;
}

// ---------------------------------------------------------------------------
// IP related stuff

// Get the local host IP address and return it in standard dot notation

char *getHostID(void)
{
	char szLclHost[120];
	LPHOSTENT lpstHostent;
	SOCKADDR_IN stLclAddr;
	SOCKADDR_IN stRmtAddr;
	int nAddrSize = sizeof(SOCKADDR);
	SOCKET hSock;
	int nRet;

	// init local address (to zero)
	stLclAddr.sin_addr.s_addr = INADDR_ANY;

	// get the local hostname
	nRet = gethostname(szLclHost, sizeof(szLclHost));
	if (nRet != SOCKET_ERROR) {
		// resolve hostname for local address
		lpstHostent = gethostbyname((LPSTR)szLclHost);
		if (lpstHostent)
			stLclAddr.sin_addr.s_addr = *((u_long FAR*) (lpstHostent->h_addr));
	}

	// if still not resolved, then try second strategy
	if (stLclAddr.sin_addr.s_addr == INADDR_ANY) {
		// get a UDP socket
		hSock = socket(AF_INET, SOCK_DGRAM, 0);
		if (hSock != INVALID_SOCKET) {
			// connect to arbitrary port and address (NOT loopback)
			stRmtAddr.sin_family = AF_INET;
			stRmtAddr.sin_port   = htons(IPPORT_ECHO);
			stRmtAddr.sin_addr.s_addr = inet_addr("160.45.10.23");
			nRet = connect(hSock, (LPSOCKADDR)&stRmtAddr, sizeof(SOCKADDR));
			if (nRet != SOCKET_ERROR) {
				// get local address
				getsockname(hSock, (LPSOCKADDR)&stLclAddr, (int FAR*)&nAddrSize);
			}
			closesocket(hSock);
		}
	}

	if (stLclAddr.sin_addr.s_addr != INADDR_ANY)
		return inet_ntoa(stLclAddr.sin_addr);

	return NULL;
}

// Return an IP address in host long notation from one supplied in standard
// dot notation

UINT4 ipstr2long(char *ip_str)
{
	char buf[6];
	char *ptr;
	int i;
	int count;
	UINT4 ipaddr;
	int cur_byte;

	ipaddr = (UINT4)0;

	for(i = 0;i < 4;i++) {
		ptr = buf;
		count = 0;
		*ptr = '\0';

		while(*ip_str != '.' && *ip_str != '\0' && count < 4) {
			if (!isdigit(*ip_str)) {
				return ((UINT4)0);
			}
			*ptr++ = *ip_str++;
			count++;
		}

		if (count >= 4 || count == 0) {
			return ((UINT4)0);
		}

		*ptr = '\0';
		cur_byte = atoi(buf);
		if (cur_byte < 0 || cur_byte > 255) {
			return ((UINT4)0);
		}

		ip_str++;
		ipaddr = ipaddr << 8 | (UINT4)cur_byte;
	}
	return (ipaddr);
}

// Check for valid IP address in standard dot notation

int good_ipaddr(char *addr)
{
	int dot_count;
	int digit_count;

	dot_count = 0;
	digit_count = 0;
	while(*addr != '\0' && *addr != ' ') {
		if (*addr == '.') {
			dot_count++;
			digit_count = 0;
		} else if (!isdigit(*addr)) {
			dot_count = 5;
		} else {
			digit_count++;
			if (digit_count > 3) {
				dot_count = 5;
			}
		}
		addr++;
	}
	if (dot_count != 3) {
		return (-1);
	} else {
		return (0);
	}
}

// Return an IP address in host long notation from a host name or address in
// dot notation

UINT4 get_ipaddr(char *host)
{
	struct hostent *hp;

	if (good_ipaddr(host) == 0) {
		return (ipstr2long(host));

	} else if ((hp = gethostbyname(host)) == (struct hostent *) NULL) {
		return ((UINT4)0);
	}

	return (ntohl(*(UINT4 *)hp->h_addr));
}

// ---------------------------------------------------------------------------
// Misc

// Do XOR of two buffers

unsigned char *xor(unsigned char *p, unsigned char *q, int length)
{
	int i;
	unsigned char *retval = p;

	for (i = 0; i < length; i++)
		*(p++) ^= *(q++);

	return retval;
}

// Return pointer to buffer of n '*'s

char *stars(int n)
{
	int i;
	static char buf[MAXPASS+1];

	for (i = 0; (i < n) && (i < MAXPASS); i++)
		buf[i] = '*';

	buf[i] = '\0';
	return &buf[0];
}

// Check whether user exists in local account database

BOOL isValidUser(LPCWSTR username)
{
	DWORD level = 0;
	LPUSER_INFO_0 pBuf = NULL;
	NET_API_STATUS stat;

	// regarding the first argument, MSDN says: "Windows NT: This string must
	// begin with \\", so maybe NULL (aka localhost) won't work on Windows NT
	stat = NetUserGetInfo(NULL, username, level, (LPBYTE *)&pBuf);

	if (stat == NERR_Success)
		return true;

	return false;
}
